"""
Utility to compile the PIVPipelineUtility.jl and main.jl scripts.
"""

import subprocess
import os
import pathlib
import logging

logging.basicConfig(level=logging.INFO, format="%(asctime)s -- %(levelname)s -- %(message)s")
logging.basicConfig(level=logging.ERROR, format="%(asctime)s -- %(levelname)s -- %(message)s")

def compile_juliapiv():
    """
    Compile juliaPIV on the target OS.
    """
    root_dir = pathlib.Path(__file__).resolve().parent   # juliaPIV/src/juliaPIV/
    pivpipelineutility_path = os.path.join(root_dir, "PIVPipelineUtility")
    precompile_dir_path = os.path.join(pivpipelineutility_path, "src/precompile")
    precompile_script_path = os.path.join(precompile_dir_path, "precompile.jl")

    cmmd = [
        'julia',
        f'--project={pivpipelineutility_path}',
        '-e',
        f"""
        using Pkg;
        Pkg.add("PackageCompiler");
        using PackageCompiler;
        Pkg.instantiate();
        create_sysimage([:PIVPipelineUtility], sysimage_path="{os.path.join(pivpipelineutility_path, "pivbuildsysimage.so")}", precompile_execution_file="{precompile_script_path}", project="{pivpipelineutility_path}");
        """
    ]
    subprocess.run(cmmd)

    print(f"Build location: {os.path.join(pivpipelineutility_path, 'pivbuildsysimage.so')}")



"""
General utilities for the PIV pipeline.
"""

from multiprocessing import cpu_count
import logging
from pathlib import Path

logging.basicConfig(level=logging.INFO, format="%(asctime)s -- %(levelname)s -- %(message)s")
logging.basicConfig(level=logging.WARNING, format="%(asctime)s -- %(levelname)s -- %(message)s")

def batch_n_nproc_logic(n, nproc, num_images):
    """
    Identify the best number of images to place in each batch depending on the values
    of N and NPROC.
    """
    settings_str = "NPROC: {}; N: {}; Number of images: {}"
    end_str = "NPROC: {}, N: {}, Batches: {}, Images per batch: {}, Total images to proces: {}"

    logging.info(f"User settings = \n\n\t{settings_str.format(nproc, n, num_images)}\n")

    if n < 2:
        logging.warning(f"N should be 2 or greater for PIV, aborting.")
        exit(1)

    max_cpus = cpu_count() - 1
    if nproc > max_cpus:
        logging.warning(f"NPROC set higher than the number of available cores: {max_cpus}")
        logging.warning(f"NPROC changed from {nproc} to {max_cpus}")
        nproc = max_cpus
    
    num_batches = find_num_batches(nproc, num_images, n)
    while num_images >= n: 
        num_batches = find_num_batches(nproc, num_images, n)
        if num_batches > 0:
            logging.info(f"Optimal settings: \n\n\t{end_str.format(num_batches, n, num_batches, num_images // num_batches, num_batches * (num_images // num_batches))}\n")
            return num_batches, num_batches,
        num_images -= 1

    logging.error(f"N = {n}, this may be more than the number of images!\n")
    raise ValueError("Could not find valid configuration given the settings.")

def find_num_batches(nproc, num_images, n):
    """
    TODO: Doc
    """
    for num_batches in range(nproc, 0, -1):
        img_p_batch = num_images // num_batches
        if img_p_batch >= n and img_p_batch % n == 0:
            return num_batches
    return 0

def build_dir_structure(parent):
    """
    Build the output directory structure.
    """
    parent = Path(parent).resolve()
    piv_mat_out = parent / "piv_mat_out"
    piv_batches = parent / "piv_batches"
    
    if not parent.is_dir():
        parent.mkdir()
    if not piv_mat_out.is_dir():
        piv_mat_out.mkdir()
    if not piv_batches.is_dir():
        piv_batches.mkdir()



"""
Author: Robin Shindelman
Date: 2025-04-09
A script to partition a directory of files into many directories with the OG
files split between them. This script was originally built to facilitate the
OSU ROXSI SVS pipeline.
"""

import os
import sys
import click
from tqdm import tqdm
from natsort import natsorted


@click.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.argument('indir')
@click.option('-o', '--out-dir', type=str, default='.',
              help='Directory to output .txt files')
@click.option('-n', '--n-files-per', type=int,
              help='Desired number of files per .txt output file, use this OR -b not both')
@click.option('-b', '--num-batches', type=int,
              help='Number of batches desired, use this OR -n not both')
def batcher_cli(indir, n_files_per, out_dir, num_batches):
    """
    A script to partition a directory of files into many directories with the OG
    files split between them. 
    """
    abs_in = os.path.abspath(indir)
    abs_out = os.path.abspath(out_dir)
    out_dir_setup(abs_out)
    generate_txt_files(abs_in, abs_out, n_files_per, num_batches)

def out_dir_setup(abs_out: str) -> None:
    """
    Ensure the out directory exists, creates it if it doesn't.
    """
    if not os.path.isdir(abs_out):
        print(f"Building a directory at {abs_out}")
        os.makedirs(abs_out)
        return
    if len(os.listdir(abs_out)) != 0:
        print("WARNING: piv_batches directory is not empty, exiting.")
        sys.exit()

def generate_txt_files(abs_in: str, abs_out: str, 
                       n_files_per: int = None,
                       num_batches: int = None
                       ) -> None:
    """
    Generate a number of .txt files based on how many individual files need to
    be parsed and how many files the user wants in each .txt file. Absolute paths
    to each file are built. 

    f-in-txt = f-total // n-files-per 
    """
    sorted_files = natsorted(os.listdir(abs_in))
    files = [os.path.join(abs_in, f) for f in sorted_files]   # Absolute after sort
    num_f = len(files)

    if num_batches:
        f_per_batch = num_f // num_batches
        leftover = num_f % num_batches
        click.echo(f"Batches: {num_batches}; Files Per Batch: {f_per_batch}; Leftover: {leftover}; From {num_f} Files")

        with tqdm(total=num_f, desc="Batching") as pbar:
            chunk = 1
            i = 0
            while i < num_f and chunk <= num_batches:
                batch_fp = os.path.join(abs_out, f'batch{chunk}.txt')
                with open(batch_fp, 'a') as f:
                    f.write(f'{files[i]}\n')

                i += 1
                pbar.update(1)
                if i == (f_per_batch * chunk):  # Makes a new batch every n files
                    chunk += 1

    elif n_files_per:   # TODO: make this available
        click.echo("Dividing by desired number of files per batch")
        return
    else:
        click.echo("No batch number or file number defined. Aborting!")
    return



"""
Top layer
"""

from .julia_compiler import compile_juliapiv
from .pivpipe.pivpipe import pivpipe_main, load_config
from .batcher.batcher import out_dir_setup, generate_txt_files
from .utils import batch_n_nproc_logic, build_dir_structure
import click
import os
import logging
from pathlib import Path
import yaml

logging.basicConfig(level=logging.INFO, format="%(asctime)s -- %(levelname)s -- %(message)s")

@click.group()
def juliaPIV():
    """
    Main entry for juliaPIV
    """
    pass

@juliaPIV.command()
def init():
    """
    Initialize the PIVPipelineUtility project. Compiling it into a full executable.
    """ 
    print("This may take a while...")
    compile_juliapiv()

@juliaPIV.command()
@click.option('--config', '-c', type=click.Path(exists=True), help="YAML config file to load defaults.")
def pipeline(config):
    """
    Run PIV pipeline.
    """
    settings = load_config(config)
    num_images = len(os.listdir(settings['input']))
    nproc, num_batches = batch_n_nproc_logic(settings['N'], settings["NPROC"], num_images)
    settings['NPROC'] = nproc

    logging.info("Building directory structure...")
    build_dir_structure(settings["output"])

    logging.info("Batching...")
    batches_dir = str(Path(settings["output"]).resolve() / "piv_batches")
    mat_dir = str(Path(settings["output"]).resolve() / "piv_mat_out")
    out_dir_setup(batches_dir)
    generate_txt_files(settings["input"], batches_dir, num_batches=num_batches)

    settings["input"], settings["output"] = batches_dir, mat_dir
    pivpipe_main(settings)
